use clap::Parser;

use rosc::encoder;
use rosc::{OscMessage, OscPacket, OscType};

use std::io::{self, Write};
use std::net::{IpAddr, Ipv6Addr, TcpListener};
use std::thread;
use std::time::Duration;

const BAUD_RATE: u32 = 115_200;

const BREAK_MICROS: u64 = 88;

#[derive(Debug, Parser)]
struct Args {
    #[arg(short, long, conflicts_with = "port")]
    list_ports: bool,

    #[arg(short = 'h', long, default_value_t = IpAddr::V6(Ipv6Addr::LOCALHOST), conflicts_with = "list_ports")]
    osc_host: IpAddr,

    #[arg(
        short = 'p',
        long,
        default_value_t = 9996,
        conflicts_with = "list_ports"
    )]
    osc_port: u16,

    #[arg(required_unless_present = "list_ports")]
    port: Option<String>,
}

fn main() {
    let args = Args::parse();

    if args.list_ports {
        for port in serialport::available_ports().unwrap() {
            println!("{}", port.port_name);
        }

        return;
    }

    let listener = TcpListener::bind((args.osc_host, args.osc_port)).unwrap();
    let (mut osc, _) = listener.accept().unwrap();

    let mut serial = serialport::new(args.port.unwrap(), BAUD_RATE)
        .open()
        .unwrap();

    let mut color = [0; 3];
    let mut strobe = false;
    loop {
        match serial.read_exact(&mut color) {
            Ok(()) => {
                if color[0] == 0 && color[1] == 0 && color[2] == 0 {
                    continue;
                }

                println!("{:#?}", color);

                let _ = send_value(&mut osc, "dmx_red".into(), color[0] as f32 / 255.0);
                let _ = send_value(&mut osc, "dmx_green".into(), color[1] as f32 / 255.0);
                let _ = send_value(&mut osc, "dmx_blue".into(), color[2] as f32 / 255.0);

                if color[0] == 255 && color[1] == 255 && color[2] == 255 {
                    if !strobe {
                        let _ = send_value(&mut osc, "dmx_strobe".into(), true);
                    }

                    strobe = true;
                } else {
                    if strobe {
                        let _ = send_value(&mut osc, "dmx_strobe".into(), false);
                    }

                    strobe = false;
                }
            }
            Err(_) => {
                thread::sleep(Duration::from_micros(BREAK_MICROS));
            }
        }
    }
}

fn send_value<W, T>(w: &mut W, addr: String, v: T) -> io::Result<()>
where
    W: Write,
    T: Into<OscType>,
{
    let buf = encoder::encode_tcp(&OscPacket::Message(OscMessage {
        addr,
        args: vec![v.into()],
    }))
    .unwrap();

    w.write_all(&buf[..])?;
    w.flush()
}
